/* fichier: "runtime.c" */

#include <stdio.h>
#include <stdlib.h>

/*
Ce module implante le runtime d'un compilateur Scheme generant du
code assembleur pour MIPS.  Le runtime s'occupe de l'initialisation du
systeme, la gestion de la memoire (allocation du tas et de la pile, et
GC du tas) et l'impression des objets.  C'est le runtime qui lance
l'execution du code assembleur genere par le compilateur.

En supposant que le compilateur Scheme a genere le programme objet
"program.s", le programme executable "a.out" est produit avec la
sequence de commandes suivantes:

  cc -g -c runtime.c
  as -g -o program.o program.s
  cc runtime.o program.o

Le GC implante ici exige que la representation suivante des objets
soit respectee TRES PRECISEMENT.  Il y a 5 types d'objets: les fixnums
(entiers de 30 bits), les paires, les fermetures, les booleens, et la
liste vide.  Chaque objet est encode par un entier de 32 bits (le type
C "word").  Pour les paires et fermetures, dans l'entier est encode
l'adresse d'un espace en memoire qui contient le reste de
l'information associee a l'objet.  Il est primordial que cette adresse
soit un multiple de 4.  Voici les details d'encodage:

  Objet           Encodage   Encodage en binaire

  entier n           n*4     NN...NNNNN00

  #f                 -2      11...1111110

  #t                 -6      11...1111010

  ()                 -10     11...1110110

  paire (car . cdr)  a+1     AA...AAAAA01   En memoire a l'adresse
                                            AA...AAAAA00: encodage(car)
                                                          encodage(cdr)

  fermeture avec     a+3     AA...AAAAA11   En memoire a l'adresse
  n variables libres                        AA...AAAAA00: n*4+2
  non-globales                                            adresse du code
                                                          encodage(var1)
                                                          ...
                                                          encodage(varn)
*/

/* Quelques macros pour encoder et decoder les objets: */

#define TAG(x)((x)&3)
#define TAG_FIXNUM  0
#define TAG_PAIR    1
#define TAG_SPECIAL 2
#define TAG_CLOSURE 3

#define OBJ_FALSE    (-1*4+TAG_SPECIAL)
#define OBJ_TRUE     (-2*4+TAG_SPECIAL)
#define OBJ_NULL     (-3*4+TAG_SPECIAL)
#define BROKEN_HEART (-4*4+TAG_SPECIAL)

#define CAR    0
#define CDR    1
#define HEADER 0
#define CODE   1

/* Definition du tas et de la pile: */

#define STACK_SIZE_IN_WORDS     1000
#define SEMISPACE_SIZE_IN_WORDS 1000

typedef long word; /* Un mot machine est un entier de 32 bits.               */

word *stack;       /* La pile est un tableau de mots.                        */
word *stack_alloc; /* Le pointeur d'allocation de la pile.                   */
word *heap;        /* Le tas est constitue de 2 semi-espaces de meme taille. */
word *heap_alloc;  /* Le pointeur d'allocation du tas.                       */
word *heap_limit;  /* La limite d'allocation du tas.                         */

/* Variables et procedures importees du code assembleur genere: */

extern word globals[];    /* La table des variables globales.   */
extern word nb_globals;   /* Le nombre de variables globales.   */
extern void start (void); /* Point d'entree du code assembleur. */

/* Programme principal: */

void main (void)
{
  /* Allouer la pile: */

  stack = (word *) malloc (STACK_SIZE_IN_WORDS*sizeof(word));
  if (stack == NULL) { printf ("ERROR -- Can't allocate stack\n"); exit (1); }

  stack_alloc = stack;

  /* Allouer le tas: */

  heap = (word *) malloc (2*SEMISPACE_SIZE_IN_WORDS*sizeof(word));
  if (heap == NULL) { printf ("ERROR -- Can't allocate heap\n"); exit (1); }

  heap_alloc = heap;
  heap_limit = heap+SEMISPACE_SIZE_IN_WORDS;

  /* Initialiser les variables globales: */

  {
    word i;
    for (i=0; i<nb_globals; i++)
      globals[i] = OBJ_FALSE;
  }

  /* Demarrer l'execution du code assembleur genere: */

  start ();
}

/* Impression des objets. */

void write_obj (word obj)
{
  switch (TAG(obj))
    {
    case TAG_FIXNUM:
      printf ("%d", obj/4);
      break;

    case TAG_SPECIAL:
      switch (obj)
	{
        case OBJ_FALSE: printf ("#f");        break;
        case OBJ_TRUE:  printf ("#t");        break;
        case OBJ_NULL:  printf ("()");        break;
	default:        printf ("{UNKNOWN}"); break;
	}
      break;

    case TAG_PAIR:
      {
        word tail = ((word *)(obj-TAG_PAIR))[CDR];
        printf ("(");
        write_obj (((word *)(obj-TAG_PAIR))[CAR]);
        while (TAG(tail) == TAG_PAIR)
          {
            printf (" ");
            write_obj (((word *)(tail-TAG_PAIR))[CAR]);
            tail = ((word *)(tail-TAG_PAIR))[CDR];
          }
        if (tail != OBJ_NULL)
          {
            printf (" . ");
            write_obj (tail);
          }
        printf (")");
        break;
      }

    case TAG_CLOSURE:
      printf ("{CLOSURE}");
      break;
    }
}

void write_obj_nl (word obj)
{
  write_obj (obj);
  printf ("\n");
  fflush (stdout);
}

/*
Le GC employe ici est du type "stop-and-copy" avec 2 semi-espaces
utilisant l'algorithme de copie en largeur de Cheney.  Le code
assembleur doit appeler la procedure "gc" lorsque le semi-espace
courant est epuise.  L'argument de cette procedure correspond au
nombre de mots qu'on cherche a allouer.  Si le GC n'arrive pas a
recuperer suffisamment d'espace pour allouer ce nombre de mots, une
erreur fatale est signalee.  A l'appel de la procedure "gc", il faut
que la variable "stack_alloc" contienne le pointeur de dessus de pile.
Au retour, la variable "heap_alloc" contient le nouveau pointeur
d'allocation du tas et "heap_limit" le nouveau pointeur limite
d'allocation du tas.

Donc, si on decide de reserver le registre 16 pour le pointeur
d'allocation de pile, il faut utiliser la sequence d'instructions
assembleur suivante pour s'assurer qu'il y a au moins 100 mots de
libre dans le tas:

     lw   $17,heap_alloc  # r[17] = heap_alloc
     lw   $18,heap_limit  # r[18] = heap_limit
     sub  $17,$18,$17     # r[17] = heap_limit - heap_alloc
     sub  $17,$17,400     # r[17] = (heap_limit - heap_alloc) - 400
     bgez $17,ok          # si r[17] >= 0, il y a assez d'espace libre
     sw   $16,stack_alloc # mettre "stack_alloc" a jour
     li   $4,100          # preparer l'argument de l'appel du "gc"
     jal  gc              # appel du "gc"
ok:

Des sequences plus rapides existent, mais la sequence precedente a
l'avantage d'etre simple a comprendre.
*/

word *heap_scan;  /* Pointeur pour balayer les objets dans le "tospace". */

/* La procedure "move" deplace un objet du "fromspace" au "tospace".  */
/* De plus, si l'objet en question a ete deplace, le pointeur est mis */
/* a jour pour refleter la nouvelle position de l'objet.              */

void move (word *ptr)
{
  word obj = *ptr;

  switch (TAG(obj))
    {
    case TAG_PAIR:
      {
        word *pair = (word *)(obj-TAG_PAIR);
        if (pair[CAR] == BROKEN_HEART)  /* L'objet a-t'il ete deplace? */
          *ptr = pair[CDR];  /* Oui, utiliser le forwarding ptr. */
        else                 /* Sinon, copier l'objet dans le "tospace". */
          {
            heap_alloc[0] = pair[0];
            heap_alloc[1] = pair[1];
            pair[CAR] = BROKEN_HEART;
            pair[CDR] = TAG_PAIR+(word)heap_alloc;
            *ptr = TAG_PAIR+(word)heap_alloc;
            heap_alloc += 2;
          }
        break;
      }
    case TAG_CLOSURE:
      {
        word *closure = (word *)(obj-TAG_CLOSURE);
        if (closure[HEADER] == BROKEN_HEART)  /* L'objet a-t'il ete deplace? */
          *ptr = closure[CODE];  /* Oui, utiliser le forwarding ptr. */
        else                     /* Sinon, copier l'objet dans le "tospace". */
          {
            word i, n = closure[HEADER]/4;
            heap_alloc[0] = closure[0];
            heap_alloc[1] = closure[1];
            for (i=1; i<=n; i++) heap_alloc[i+1] = closure[i+1];
            closure[HEADER] = BROKEN_HEART;
            closure[CODE]   = TAG_CLOSURE+(word)heap_alloc;
            *ptr = TAG_CLOSURE+(word)heap_alloc;
            heap_alloc += n+2;
          }
        break;
      }
    }
}

void move_reachable_from_moved (void)
{
  while (heap_scan < heap_alloc)
    {
      word x = *heap_scan;
      if ((TAG(x) == TAG_SPECIAL) && (x >= 0)) /* paire ou fermeture? */
        { /* fermeture! */
          word i, n = x/4;
          for (i=1; i<=n; i++) move (&heap_scan[i+1]);
          heap_scan += n+2;
        }
      else
        { /* paire! */
          move (&heap_scan[CAR]);
          move (&heap_scan[CDR]);
          heap_scan += 2;
        }
    }
}

void move_reachable_from_roots (void)
{
  word i;

  /* Deplacer les objets contenus dans les variables globales: */

  for (i=nb_globals-1; i>=0; i--)
    move (&globals[i]);

  /* Deplacer les objets contenus dans la pile: */

  for (i=(stack_alloc-stack)-1; i>=0; i--)
    move (&stack[i]);
}

void gc (word nb_words_needed)
{
  printf ("GC --");
  fflush (stdout);

  if (heap_limit == heap+SEMISPACE_SIZE_IN_WORDS)
    heap_alloc = heap+SEMISPACE_SIZE_IN_WORDS;
  else
    heap_alloc = heap;

  heap_limit = heap_alloc+SEMISPACE_SIZE_IN_WORDS;
  heap_scan = heap_alloc;

  move_reachable_from_roots ();
  move_reachable_from_moved ();

  printf (" %d%% of semispace is live\n",
          100*(SEMISPACE_SIZE_IN_WORDS-(heap_limit-heap_alloc))
          /SEMISPACE_SIZE_IN_WORDS);
  fflush (stdout);

  if (heap_limit-heap_alloc < nb_words_needed)
    {
      printf ("ERROR -- Heap exhausted\n");
      exit (1);
    }
}
